#!/usr/bin/env python

##-----------------------------------------------------------------------------
## Title      : Driver example
## Project    : FMC DEL 1ns 2cha (FFPG)
## URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
##-----------------------------------------------------------------------------
## File       : FFPG_driver.py
## Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
## Company    : CERN (BE-BI-QP)
## Created    : 2016-08-09
## Last update: 2016-08-24
## Standard   : Python
##-----------------------------------------------------------------------------
## Description: Example functions how to operate the FFPG gateware via its
##              Wishbone interface.
##-----------------------------------------------------------------------------
## Copyright (c) 2016 CERN (BE-BI-QP)
##-----------------------------------------------------------------------------
## GNU LESSER GENERAL PUBLIC LICENSE
##-----------------------------------------------------------------------------
## This source file is free software; you can redistribute it and/or modify it
## under the terms of the GNU Lesser General Public License as published by the
## Free Software Foundation; either version 2.1 of the License, or (at your
## option) any later version. This source is distributed in the hope that it
## will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
## of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU Lesser General Public License for more details. You should have
## received a copy of the GNU Lesser General Public License along with this
## source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
##-----------------------------------------------------------------------------
## Revisions  :
## Date        Version  Author        Comment
## 2016-08-24  1.0      Jan Pospisil
## 2016-09-01  1.1      Jan Pospisil  added version numbers (major.minor.rev)
##                                    added clock_stable status bit
##                                    added option for negative pulses
## 2016-09-02  1.2      Michael Betz  added reset of MC100EP140 bad state;
##                                    better configuration; better debugging;
##                                    comments; attempt for better trigger
##                                    latency calibration (not working yet)
##----------------------------------------------------------------------------- 

# TODO: proper BUSY bits checking

import sys
import time
sys.path.append('/usr/local/encore')
import encoreio

class ConColors:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m' # better readable with BOLD
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    NONE = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    
m = encoreio.Module.get_instance('fmc_fpg', 0)

WB_DEBUG = False
WB_DEBUG_2_SV = False
WB_DEBUG_2_SV_FILE = 'wb_trace.svh'

###################################################################
## debug preparation
###################################################################

if WB_DEBUG_2_SV:
    WB_DEBUG_2_SV_FP = open(WB_DEBUG_2_SV_FILE, 'w')
    WB_DEBUG_2_SV_FP.write('uint32_t dataOut;\n')
    WB_DEBUG_2_SV_FP.write('int errors = 0;\n')
    WB_DEBUG_2_SV_FP.write('\n')

###################################################################
## WishBone access overlay
###################################################################

def WbRead(register):
    result = m.read(register)[0]
    if WB_DEBUG:
        print(ConColors.MAGENTA+'WB_READ:'+ConColors.NONE+' ['+register+'] = '+hex(result))
    if WB_DEBUG_2_SV:
        address = int(m[register]['offset'], 16)
        WB_DEBUG_2_SV_FP.write('WbRead('+str(address)+', dataOut, "WB_READ: '+register+':");\n')
        WB_DEBUG_2_SV_FP.write('assert (dataOut == '+str(result)+') else begin\n')
        WB_DEBUG_2_SV_FP.write('    $error("WB_READ different, result should be '+hex(int(result))+'!");\n')
        WB_DEBUG_2_SV_FP.write('    errors = errors + 1;\n')
        WB_DEBUG_2_SV_FP.write('end\n')
    return result

def WbWrite(register, data):
    if WB_DEBUG:
        print(ConColors.CYAN+'WB_WRITE:'+ConColors.NONE+' ['+register+'] <- '+hex(data))
    if WB_DEBUG_2_SV:
        address = int(m[register]['offset'], 16)
        WB_DEBUG_2_SV_FP.write('WbWrite('+str(address)+', '+str(data)+', "WB_WRITE: '+register+':");\n')
    m.write(register, (data,))

def WbSetBits(register, mask, bits = -1):
    if bits == -1:
        bits = mask
    value = WbRead(register)
    value &= (~mask)
    value |= (bits & mask)
    WbWrite(register, value)

def WbReadMulti(register, start = 0, to = -1):
    result = m.read(register, start=start, to=to)
    if WB_DEBUG:
        print(ConColors.MAGENTA+'WB_READ_MULTI:'+ConColors.NONE+' ['+register+']('+str(start)+':'+str(to)+') = ')
        hexDump( result )
    if WB_DEBUG_2_SV:
        address = int(m[register]['offset'], 16)
        if to == -1:
            to = int(m[register]['depth'], 16)
        for i in range(start, to):
            WB_DEBUG_2_SV_FP.write('WbRead('+str(address+i*4)+', dataOut, "WB_READ_MULTI: '+register+'['+str(i)+']:");\n')
            WB_DEBUG_2_SV_FP.write('assert (dataOut == '+str(result[i-start])+') else begin\n')
            WB_DEBUG_2_SV_FP.write('    $error("WB_READ_MULTI different, result should be '+hex(int(result[i-start]))+'!");\n')
            WB_DEBUG_2_SV_FP.write('    errors = errors + 1;\n')
            WB_DEBUG_2_SV_FP.write('end\n')
    return result

def WbWriteMulti(register, data, start = 0, to = -1):
    if WB_DEBUG:
        print(ConColors.CYAN+'WB_WRITE_MULTI:'+ConColors.NONE+' ['+register+']('+str(start)+':'+str(to)+') <- ')
        hexDump( data )
    if WB_DEBUG_2_SV:
        address = int(m[register]['offset'], 16)
        if to == -1:
            to = int(m[register]['depth'])
        for i in range(start, to):
            WB_DEBUG_2_SV_FP.write('WbWrite('+str(address+i*4)+', '+str(data[i-start])+', "WB_WRITE_MULTI: '+register+'['+str(i)+']:");\n')
    m.write(register, data, start=start, to=to)

def WbSetBitsMulti(register, mask, bits, start = 0, to = -1):
    if len(mask) != len(bits):
        raise Exception('Different MASK and BITS length!')
    values = WbReadMulti(register, start, to)
    values = list(values)
    if len(values) != len(mask):
        raise Exception('Different length of MASK/BITS and actual data!');
    for i in range(len(values)):
        values[i] &= (~mask[i])
        values[i] |= (bits[i] & mask[i])
    WbWriteMulti(register, values, start, to)

# timeOut in [s] (floating point)
# polarity = 1: (register.bit == 1) <=> BUSY, wait for 0
# polarity = 0: (register.bit == 0) <=> BUSY, wait for 1
def WaitForNotBusy(register, bit, timeOut, tries, message = '', polarity = 1):
    busy = True
    for i in range(tries):
        if (WbRead(register) & (1<<bit)) != polarity:
            busy = False
            break
        # wait for timeOut seconds
        timeToWait = timeOut
        while timeToWait > 0:
            timeBegin = time.time()
            time.sleep(timeToWait)
            timeToWait -= time.time() - timeBegin
    if busy:
        raise Exception('WaitForNotBusy error: '+message)

###################################################################
## FFPG specific functions
###################################################################

# synchronize AD9512 dividers
def Ad9512Sync():
    WbSetBits('control', 0x200)

def PrintFrequency():
    print(str(WbRead('frequency')/1e6)+' MHz')

# 0 - front panel clock
# 1 - FPGA loop clock
# 2 - oscillator
def SelectClock(clock):
    if (clock < 0) or (clock > 2):
        raise Exception("Unknown clock!")
    WbSetBits('control', 3, clock)

# ratio 1-32
def SetRatio(ratio):
    if (ratio < 1) or (ratio > 32):
        raise Exception("Bad ratio!")
    WbWrite('clock_ratio_m1', ratio-1)

# voltage in [V]
def SetTriggerThreshold(voltage):
    vdd = 5
    bits = 16
    code = int((voltage/float(vdd)) * (2**bits))
    code &= (2**bits)-1
    WbWrite('trig_threshold', code)

# compensation <-1;1)
def SetVcxoFrequency(compensation):
    vdd = 3.3
    bits = 16
    voltage = vdd * (compensation + 1) / 2.0
    code = int((voltage/float(5)) * (2**bits))
    code &= (2**bits)-1
    WbWrite('vcxo_voltage', code)

# channel: 1, 2, 0 (for both)
# memory: 'set', 'res', '*'
# if memoryPart > -1: clear only first memoryPart words
def ClearMemory(channel, memory, memoryPart = -1):
    if channel == 1:
        address = 'ch1_'
    elif channel == 2:
        address = 'ch2_'
    elif channel == 0:
        ClearMemory(1, memory, memoryPart)
        ClearMemory(2, memory, memoryPart)
        return
    else:
        raise Exception("Bad channel!")
    
    if memory == "set":
        address += 'set_mem'
    elif memory == "res":
        address += 'res_mem'
    elif memory == "*":
        ClearMemory(channel, 'set', memoryPart)
        ClearMemory(channel, 'res', memoryPart)
        return
    else:
        raise Exception("Bad memory!")
    
    partToClear = 2048
    to = -1
    if memoryPart > -1:
        partToClear = memoryPart
        to = memoryPart
    WbWriteMulti(address, (0,)*partToClear, to = to)

def getRegname( channel, targ ):
    """ channel: 1, 2    targ: "SET", "RESET" """
    regName = {"SET":"ch%1d_set_mem", "RESET":"ch%1d_res_mem"}[targ]%channel
    return regName

def putBit( channel, ind, val, targ="SET" ):
    """ 
    channel: 1, 2, 0 (for both)
    enable (val=1) or disable (val=0) 
    a bit in the "SET" or "RESET" array at index `ind`.
    """
    regName = getRegname( channel, targ )
    # The pulse generator arrays is made up of 32 bit words
    # Here we split the array index into a memory address and a bit-index [0 - 31]
    val = val & 0x01
    position = ind/32
    shift = ind%32
    WbSetBitsMulti( regName, ((1<<shift),), ((val<<shift),), start=position, to=position+1)


def tobin(x, count=8): 
    """ Integer to binary. Count is number of padding bits """ 
    return "".join(map(lambda y:str((x>>y)&1), range(count-1, -1, -1)))
    
def tohex(x, count=8):
    """ Integer to hex. Count is number of padding digits """ 
    return "".join(map(lambda y:str((x>>y)&1), range(count-1, -1, -1)))

def hexDump( dataArray ):
    """ the ultimate hexdump function """
    i = 0
    for word in dataArray:
        if i%8 == 0:
            if i != 0:
                sys.stdout.write("\n")
            sys.stdout.write( "%4x: "%(i*4) )
        #sys.stdout.write( tobin(word,32) + " " )  #binary format
        sys.stdout.write( "%08x "%word )           #hex format
        i += 1
    sys.stdout.write("\n")

def ConfigurePulse(channel, start, stop, pulse, which = 0, polarity = 1):
    """
     Manipulate the SET and RESET pulse generator arrays:
        
     channel: 1, 2, 0 (for both)
     pulse: 0 - delete, 1 - create
     which: 0 - both, 1 - set, 2 - reset
     polarity: 0 - negative pulse, 1 - positive pulse

                    |             which: 0             |
                    |    which: 1    |    which: 2     |
        pulse: 0    | SET[start] = 0 | RESET[stop] = 0 |
        pulse: 1    | SET[start] = 1 | RESET[stop] = 1 |
        
    """
    if channel == 1:
        address = 'ch1_'
    elif channel == 2:
        address = 'ch2_'
    elif channel == 0:
        ConfigurePulse(1, start, stop, pulse, which, polarity)
        ConfigurePulse(2, start, stop, pulse, which, polarity)
        return
    else:
        raise Exception("Bad channel!")
    
    if (start < 0) or (start > 65535):
        raise Exception("Bad start!")
    if (stop < 0) or (stop > 65535):
        raise Exception("Bad stop!")
    if (stop < start):
        raise Exception("Bad order!")
    
    pulse &= 1          # Force all bits to zero except the first one
    
    if polarity == 0:   # Swap `start` and `stop` for inverted mode
        tmp = start
        start = stop
        stop = tmp
    
    # The pulse generator arrays is made up of 32 bit words
    # Here we split the array index into a memory address and a bit-index [0 - 31]
    if which < 2:
        position = start/32
        shift = start%32
        WbSetBitsMulti(address+'set_mem', ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

    if which != 1:
        position = stop/32
        shift = stop%32
        WbSetBitsMulti(address+'res_mem', ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

def CreatePulse(channel, start, stop, which = 0, polarity = 1):
    ConfigurePulse(channel, start, stop, 1, which, polarity)

def DeletePulse(channel, start, stop):
    ConfigurePulse(channel, start, stop, 0)

def PrintBit(value, bit, label, ifSet, ifNotSet):
    print('  '+'['+str(bit)+'] '+label+':')
    if value & (1<<bit):
        print(ConColors.GREEN+'    1: '+ifSet+ConColors.NONE)
    else:
        print(ConColors.RED+'    0: '+ifNotSet+ConColors.NONE)

def PrintBits(value, bitFrom, bitTo, label, descriptions):
    print('  '+'['+str(bitTo)+':'+str(bitFrom)+'] '+label+':')

    value &= (1<<bitTo+1)-1
    value >>= bitFrom

    color = ConColors.YELLOW
    description = descriptions[value]
    if value > len(descriptions)-1:
        color = ConColors.RED
        description = '<unknown>'

    print(color+'    '+str(value)+': '+description+ConColors.NONE)
        
def Status():
    print('Status register:')
    status = WbRead('status')
    PrintBit(status, 0, 'Clock infrastructure configuration', 'busy', 'idle')
    PrintBit(status, 1, 'VCXO DAC', 'busy', 'idle')
    PrintBit(status, 2, 'Trigger threshold DAC', 'busy', 'idle')
    PrintBit(status, 3, 'Delay configuration', 'busy', 'idle')
    PrintBit(status, 4, 'Channel 1 output', 'enabled', 'disabled')
    PrintBit(status, 5, 'Channel 2 output', 'enabled', 'disabled')
    PrintBit(status, 6, 'Channel 1', 'running', 'stopped')
    PrintBit(status, 7, 'Channel 2', 'running', 'stopped')
    PrintBit(status, 8, 'Input clock stability', 'stable', 'not stable/present')

def Control():
    print('Control register: ')
    control = WbRead('control')
    # PrintBits(control, 0, 1, 'Clock selection', ('external clock', 'FPGA LOOP', 'onboard VCXO')) # not implemented
    PrintBit(control, 0, 'Clock mux', 'VCXO', 'FPGA LOOP')
    PrintBit(control, 2, 'Channel 1 output', 'enabled', 'disabled')
    PrintBit(control, 3, 'Channel 2 output', 'enabled', 'disabled')
    PrintBits(control, 4, 5, 'Channel 1 mode', ('stopped', 'continuous', 'one-shot'))
    PrintBits(control, 6, 7, 'Channel 2 mode', ('stopped', 'continuous', 'one-shot'))
    PrintBit(control, 8, 'LEDs', 'blinking', 'normal operation')
    PrintBit(control, 9, 'AD9512 Synchronization', 'in progress', 'done')

def Enable(channel):
    channel -= 1
    channel %= 2
    WbSetBits('control', 0x4<<channel, 0x4<<channel) # set output enable

def Disable(channel):
    channel -= 1
    channel %= 2
    WbSetBits('control', 0x4<<channel, 0) # set output disable

def Start(channel):
    channel -= 1
    channel %= 2
    channel *= 2
    WbSetBits('control', 0x30<<channel, 0x10<<channel) # set mode CONT.

def StartOnce(channel):
    channel -= 1
    channel %= 2
    channel *= 2
    WbSetBits('control', 0x30<<channel, 0x20<<channel) # set mode ONE-SHOT

def Stop(channel):
    channel -= 1
    channel %= 2
    channel *= 2
    WbSetBits('control', 0x30<<channel, 0) # set mode STOP

def ResetBadState(channel):
    """ robust reset function to get out of the bad `state 1` """
    Disable( channel )                  #Outputs Hi-Z
    Stop( channel )                     #Stop clock
    ClearMemory( channel, '*' )         #Clear the pulse generator arrays
    # Define a single SET pulse per cycle, 
    # after two cycles, the MC100EP140 should be in the good
    # state number 3 (output high)
    ConfigurePulse(channel, 10, 50, 1, 1, 1)
    Start( channel )                    #Enable clock
    time.sleep( 1e-3 )
    Stop( channel )
    ClearMemory( channel, '*' )

def Debug():
    print('Debug:')
    debug = WbRead('debug')
    PrintBits(debug, 0, 2, "CH1 FSM state", ("Stop", "WaitForTrigger", "Generating", "Outputting"))
    PrintBits(debug, 3, 5, "CH2 FSM state", ("Stop", "WaitForTrigger", "Generating", "Outputting"))

def Version():
    version = WbRead('version')
    major = (version >> 22) & ((2**10)-1)
    minor = (version >> 12) & ((2**10)-1)
    revision = version & ((2**12)-1)
    print('Gateware version: '+str(major)+'.'+str(minor)+'.'+str(revision))

###################################################################
## OneWire stuff - not working yet
###################################################################

OW_CMD_ROM_SEARCH = 0xF0
OW_CMD_ROM_READ = 0x33
OW_CMD_ROM_MATCH = 0x55
OW_CMD_ROM_SKIP = 0xCC
OW_CMD_ALARM_SEARCH = 0xEC
OW_CMD_CONVERT_T = 0x44
OW_CMD_WRITE_SCRATCHPAD = 0x4E
OW_CMD_READ_SCRATCHPAD = 0xBE
OW_CMD_COPY_SCRATCHPAD = 0x48
OW_CMD_RECALL_E_2 = 0xB8
OW_CMD_READ_POWER_SUPPLY = 0xB4

# overdrive: 0 - normal mode / 1 - overdrive mode

def OneWireReset(overdrive = 0):
    print 'OW: reset'
    WbWrite('onewire_ctrl_sta', 0x08 | (overdrive << 2) | 0x02)
    reg = WbRead('onewire_ctrl_sta')
    while (0x08 & reg):
        reg = WbRead('onewire_ctrl_sta')
    reg &= 0x01
    print 'OW: presence = '+str(reg)
    return reg

def OneWireSlot(bit, overdrive = 0):
    WbWrite('onewire_ctrl_sta', 0x08 | (overdrive << 2) | bit)
    reg = WbRead('onewire_ctrl_sta')
    while (0x08 & reg):
        reg = WbRead('onewire_ctrl_sta')
    return reg & 0x01

def OneWireReadBit(overdrive = 0):
    result = OneWireSlot(overdrive, 1)
    print 'OW: read bit = '+str(result)
    return result

def OneWireWriteBit(bit, overdrive = 0):
    print 'OW: write bit('+str(bit)+')'
    OneWireSlot(overdrive, bit)

def OneWireRead(overdrive = 0):
    result = 0;
    for i in range(8):
        result |= OneWireReadBit(overdrive)
        result <<= 1
    print 'OW: read = '+hex(result)
    return result

def OneWireWrite(byte, overdrive = 0):
    print 'OW: write('+hex(byte)+')'
    for i in range(8):
        OneWireWriteBit(byte & 1, overdrive)
        byte >>= 1

def OneWireReadRom(overdrive = 0):
    presence = OneWireReset(overdrive)
    if presence == 0:
        print 'OW Thermometer is not present!'
        return
    OneWireWrite(OW_CMD_ROM_READ, overdrive)
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())
    print hex(OneWireRead())

###################################################################
## Test functions
###################################################################

def Init(memoryPart = -1):
    WbWrite('control', 0)
    SelectClock(0)
    SetRatio(2)
    Ad9512Sync()
    SetTriggerThreshold(0.5)
    SetVcxoFrequency(-0.2105) # 125.0000 MHz
    ClearMemory(0, '*', memoryPart)
    xx = 56.28 # problem mazbe in CalculateParams
    print(xx)
    Calibrate(xx) # 63.531 in [ns], depends on setup, cable length...

# how much is RF clock faster than bunch clock
RF_2_BUNCH_RATIO = 5

globalFineOffset = 0

# time 't' is in [ns] - same as other times
# output: (bit, delay)
#   bit - position of the bit to set
#   delay - delay parameter to be set
def CalculateParams(t, intraBunchCheck = 0):
    global globalFineOffset
    rfPeriod = 4.99 # 1/Clk_rf
    delayOffset = globalFineOffset #3.714 # fine compensation of trigger delay
    delayScale = 1.0362 # non-linearity of delay IC
    
    if (intraBunchCheck) and ((t < 0) or (t > RF_2_BUNCH_RATIO * rfPeriod)):
        raise Exception("Parameter t can be just in one bunch clock period!")
    
    bit = int(t/rfPeriod) # bit position
    rest = t - bit * rfPeriod # in ns
    value = int(round(delayScale * rest * 100))
    value += delayOffset
    delay = value # delay IC value (10 ps step)
    
    return (bit, delay)
    
# delay and width (both in [ns]) of thi pulse within one bunch clock
# output: (setBit, resBit) - bit offset within one bunch clock
def SetPulseShape(channel, delay, width, intraBunchCheck = 0):
    # calculate bitt shifts and delays
    (setBit, setDelay) = CalculateParams(delay, intraBunchCheck)
    (resBit, resDelay) = CalculateParams(delay + width, intraBunchCheck)
    # set delays
    WbWrite('ch'+str(channel)+'_delay_set', setDelay)
    WbWrite('ch'+str(channel)+'_delay_res', resDelay+18) # 18 is compensating difference between set and res relay IC
    # return bit shifts
    return (setBit, resBit)
    
# pulseDelay and pulseWidth in ns
def SetPulse(channel, pulseDelay, pulseWidth, polarity = 1):
    ResetBadState(channel)
    (setBit, resBit) = SetPulseShape(channel, pulseDelay, pulseWidth)
    CreatePulse(channel, setBit, resBit, polarity = polarity)

def SetPulsesInBunches(channel, listOfBunches, pulseDelay, pulseWidth, polarity = 1):
    ResetBadState(channel)
    (setBit, resBit) = SetPulseShape(channel, pulseDelay, pulseWidth, intraBunchCheck = 1)
    for bunch in listOfBunches:
        CreatePulse(channel, (bunch * RF_2_BUNCH_RATIO)+setBit, (bunch * RF_2_BUNCH_RATIO)+resBit, polarity = polarity)

def TestPulseInBunches(channel = 1, overflow = 17820, polarity = 1):
    # configure channel
    WbWrite('overflow', overflow)

    pulseDelay = 10 # inter-bunch pulse delay in [ns]
    pulseWidth = 10 # inter-bunch pulse width in [ns]
    bunches = (0,1,2,10,50) # in which bunches pulse should be
    # bunches = range(overflow / RF_2_BUNCH_RATIO)
    # bunches = (0,)
    
    SetPulsesInBunches(channel, bunches, pulseDelay, pulseWidth)
    
    Enable(channel)
    Start(channel)

def TestPulse(channel = 1, overflow = 17820, polarity = 1):
    # configure channel
    WbWrite('overflow', overflow)

    SetPulse(channel, pulseDelay = 0, pulseWidth = 100, polarity=polarity)
    
    Enable(channel)
    Start(channel)

# phaseOffset in [ns] - real offset to the real beam
def Calibrate(phaseOffset):
    global globalFineOffset
    globalFineOffset = 0
    (bitOffset, fineOffset) = CalculateParams(phaseOffset)
    print('bitOffset = '+str(bitOffset)+', fineOffset = '+str(fineOffset))
    WbWrite('trig_latency', bitOffset)
    globalFineOffset = fineOffset

Init()
Control()
Status()

#TestPulseInBunches(1, polarity = 0)
TestPulse(1, polarity = 1)

Control()
Status()
Version()
Debug()


# OneWireReadRom(1) # does not work :/

###################################################################
## debug ending
###################################################################

if WB_DEBUG_2_SV:
    WB_DEBUG_2_SV_FP.write('\n')
    WB_DEBUG_2_SV_FP.write('$display("Error count: %0d\\n", errors);\n')
    WB_DEBUG_2_SV_FP.close()
    WB_DEBUG_2_SV = False
